#!/usr/local/bin/python
#
#    bhmm - Bayesian HMM in Python.
#    $Id: bhmm.py,v 1.4 2019/03/15 11:41:37 daichi Exp $
#
import hmm
import sys
import getopt
import numpy as np

def eprint (s):
    sys.stderr.write (s + '\n')
    sys.stderr.flush ()

def eprintf (s):
    sys.stderr.write (s)
    sys.stderr.flush ()

def listload (file):
    sys.stderr.write ('reading data from %s.. ' % file),
    sys.stderr.flush ()
    data = []
    with open (file, 'r') as fh:
        for line in fh:
            words = line.rstrip('\n').split()
            if len(words) > 0:
                data.append (map(int, words))
    sys.stderr.write ('done.\n')
    return data

def save_model (file, model):
    eprint ('saving model to %s ..' % file)
    save_zz (file, model)
    save_emit (file, model)
    save_trans (file, model)
    eprint ('done.')

def save_zz (file, model):
    with open (file + '.zz', 'w') as fh:
        for zz in model['zz']:
            fh.write(' '.join(map(str, zz)) + '\n')

def save_emit (file, model):
    np.savetxt (file + '.emit', model['emit'].T, fmt='% .7f')

def save_trans (file, model):
    np.savetxt (file + '.trans', model['trans'], fmt='% .7f')

def usage ():
    print 'bhmm.py: Bayesian HMM in Python.'
    print 'usage: % bhmm.py OPTIONS train model'
    print 'OPTIONS'
    print ' -K states  number of states in HMM'
    print ' -N iters   number of MCMC iterations'
    print ' -a alpha   Dirichlet hyperparameter on transitions (default 0.1)'
    print ' -b beta    Dirichlet hyperparameter on emissions (default 0.01)'
    print ' -h         displays this help'
    print '$Id: bhmm.py,v 1.4 2019/03/15 11:41:37 daichi Exp $'
    sys.exit (0)

def main ():
    shortopts = 'K:N:a:b:h'
    longopts = ['states=', 'iters=', 'alpha=', 'beta=', 'help']
    K     = 10
    iters = 1
    alpha = 0.1
    beta  = 0.01
    
    try:
        opts,args = getopt.getopt (sys.argv[1:], shortopts, longopts)
    except getopt.GetoptError as err:
        usage ()
        
    for o,a in opts:
        if o in ('-K', '--states'):
            K = int (a)
        elif o in ('-N', '--iters'):
            iters = int (a)
        elif o in ('-a', '--alpha'):
            alpha = float (a)
        elif o in ('-b', '--beta'):
            beta = float (a)
        elif o in ('-h', '--help'):
            usage ()
        else:
            assert False, "unknown option"

    if len(args) == 2:
        data    = listload (args[0])
        outfile = args[1]
    else:
        usage ()

    eprint('HMM: K = %d iters = %d alpha = %g beta = %g' \
            % (K,iters,alpha,beta))
            
    # main
    model = hmm.learn (data, K, alpha, beta, iters)
    save_model (outfile, model)

if __name__ == "__main__":
    main ()
